//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  fregistry.hpp
//! \brief Registration of CFunc's.


//  Low-level function types.

typedef void   (*func_0_i) (int);
typedef void   (*func_0_d) (double);
typedef void   (*func_0_s) (string);

typedef int    (*func_i_i) (int);
typedef int    (*func_i_d) (double);
typedef int    (*func_i_s) (string);
typedef double (*func_d_i) (int);
typedef double (*func_d_d) (double);
typedef double (*func_d_s) (string);
typedef void   (*func_e_e) (double&,double&, double,double);
typedef string (*func_s_s) (string);

typedef int    (*func_i_ii) (int, int);
typedef int    (*func_i_id) (int, double);
typedef int    (*func_i_di) (double, int);
typedef int    (*func_i_dd) (double, double);
typedef int    (*func_i_si) (string, int);
typedef double (*func_d_ii) (int, int);
typedef double (*func_d_id) (int, double);
typedef double (*func_d_di) (double, int);
typedef double (*func_d_dd) (double, double);
typedef double (*func_d_sd) (string, double);
typedef void   (*func_e_dd) (double&,double&, double, double);
typedef void   (*func_e_ee) (double&,double&, double,double, double,double);
typedef void   (*func_e_se) (double&,double&, string, double,double);
typedef string (*func_s_si) (string, int);
typedef string (*func_s_ss) (string, string);

typedef int    (*func_i_iii) (int, int, int);
typedef double (*func_d_iid) (int, int, double);
typedef double (*func_d_idi) (int, double, int);
typedef double (*func_d_idd) (int, double, double);
typedef double (*func_d_ddd) (double, double, double);
typedef void   (*func_e_eee) (double&,double&, double,double, double,double, double,double);


//! A singleton class, holding a collection of CFunc's.

class SFuncRegistry {
 private:
    SFuncRegistry() {};
    SFuncRegistry(SFuncRegistry const&);  //! To prevent copying; don't implement.
    void operator=(SFuncRegistry const&); //! To prevent copying; don't implement.

    map<string,class CFunc*> FMap;       //! unsorted hash, for expression evaluation
    vector<const class CFunc*> FList;    //! sorted array, for help text

    void register_fct_template(
        const char* _tag, const char* _outtype, const char* _intypes, void* _f );

public:
    static SFuncRegistry& getInstance()
        {   static SFuncRegistry instance; //! Instantiated on first use.
            return instance; }

    const class CFunc* find( string nam ) const;
    const class CFunc* find_or_fail( string nam ) const;

    void display_functions() const;

    void register_fct_meta(
        const char* _tag, int _narg, const char* _explanation, int _precedence=0 );

    void register_fct_0_i( const char* _tag, func_0_i _f )
    {    register_fct_template( _tag, "0", "i", (void*)_f ); }

    void register_fct_0_d( const char* _tag, func_0_d _f )
    {    register_fct_template( _tag, "0", "d", (void*)_f ); }

    void register_fct_0_s( const char* _tag, func_0_s _f )
    {    register_fct_template( _tag, "0", "s", (void*)_f ); }

    void register_fct_i_i( const char* _tag, func_i_i _f )
    {    register_fct_template( _tag, "i", "i", (void*)_f ); }

    void register_fct_i_d( const char* _tag, func_i_d _f )
    {    register_fct_template( _tag, "i", "d", (void*)_f ); }

    void register_fct_i_s( const char* _tag, func_i_s _f )
    {    register_fct_template( _tag, "i", "s", (void*)_f ); }

    void register_fct_d_i( const char* _tag, func_d_i _f )
    {    register_fct_template( _tag, "d", "i", (void*)_f ); }

    void register_fct_d_d( const char* _tag, func_d_d _f )
    {    register_fct_template( _tag, "d", "d", (void*)_f ); }

    void register_fct_d_s( const char* _tag, func_d_s _f )
    {    register_fct_template( _tag, "d", "s", (void*)_f ); }

    void register_fct_e_e( const char* _tag, func_e_e _f )
    {    register_fct_template( _tag, "e", "e", (void*)_f ); }

    void register_fct_i_ii( const char* _tag, func_i_ii _f )
    {    register_fct_template( _tag, "i", "ii", (void*)_f ); }

    void register_fct_i_id( const char* _tag, func_i_id _f )
    {    register_fct_template( _tag, "i", "id", (void*)_f ); }

    void register_fct_i_di( const char* _tag, func_i_di _f )
    {    register_fct_template( _tag, "i", "di", (void*)_f ); }

    void register_fct_i_dd( const char* _tag, func_i_dd _f )
    {    register_fct_template( _tag, "i", "dd", (void*)_f ); }

    void register_fct_i_si( const char* _tag, func_i_si _f )
    {    register_fct_template( _tag, "i", "si", (void*)_f ); }

    void register_fct_d_ii( const char* _tag, func_d_ii _f )
    {    register_fct_template( _tag, "d", "ii", (void*)_f ); }

    void register_fct_d_dd( const char* _tag, func_d_dd _f )
    {    register_fct_template( _tag, "d", "dd", (void*)_f ); }

    void register_fct_d_sd( const char* _tag, func_d_sd _f )
    {    register_fct_template( _tag, "d", "sd", (void*)_f ); }

    void register_fct_e_dd( const char* _tag, func_e_dd _f )
    {    register_fct_template( _tag, "e", "dd", (void*)_f ); }

    void register_fct_e_ee( const char* _tag, func_e_ee _f )
    {    register_fct_template( _tag, "e", "ee", (void*)_f ); }

    void register_fct_e_se( const char* _tag, func_e_se _f )
    {    register_fct_template( _tag, "e", "se", (void*)_f ); }

    void register_fct_s_si( const char* _tag, func_s_si _f )
    {    register_fct_template( _tag, "s", "si", (void*)_f ); }

    void register_fct_s_ss( const char* _tag, func_s_ss _f )
    {    register_fct_template( _tag, "s", "ss", (void*)_f ); }

    void register_fct_i_iii( const char* _tag, func_i_iii _f )
    {    register_fct_template( _tag, "i", "iii", (void*)_f ); }

    void register_fct_d_iid( const char* _tag, func_d_iid _f )
    {    register_fct_template( _tag, "d", "iid", (void*)_f ); }

    void register_fct_d_idi( const char* _tag, func_d_idi _f )
    {    register_fct_template( _tag, "d", "idi", (void*)_f ); }

    void register_fct_d_idd( const char* _tag, func_d_idd _f )
    {    register_fct_template( _tag, "d", "idd", (void*)_f ); }

    void register_fct_d_ddd( const char* _tag, func_d_ddd _f )
    {    register_fct_template( _tag, "d", "ddd", (void*)_f ); }

    void register_fct_e_eee( const char* _tag, func_e_eee _f )
    {    register_fct_template( _tag, "e", "eee", (void*)_f ); }

};

//! The one global instance of SFuncRegistry.

extern SFuncRegistry* GFuncRegistry;