//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 2001-                                               
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  func.hpp
//! \brief Function wrapper classes CTypedFunc and CFunc.

//! A wrapper holding a function with given input and output types.

class CTypedFunc {
 public:
    const string outtype;
    const string intypes;
    void* f; // the polymorphic function pointer
    CTypedFunc( const char* _outtype, const char* _intypes, void* _f )
        : outtype(_outtype), intypes(_intypes), f(_f) {};
};
    
//! A wrapper holding a function.

class CFunc {
 public:
    const string tag;
    int narg;
    const char* explanation;
    int precedence;
    vector<const CTypedFunc*> tyfuList; //! List of available typed function instances.
    CFunc( const char* _tag, int _narg, const char* _explanation, int _precedence=0 )
        : tag(_tag), narg(_narg), explanation(_explanation), precedence(_precedence) {};
    const CTypedFunc* find_exact_tyfu( string intypes ) const;
    const CTypedFunc* find_tyfu( string intypes, bool want_error ) const;
    class CCoord coord( class CCoord *co ) const;
    class CCoord coord( class CCoord *co1, class CCoord *co2 ) const;
    class CCoord coord( class CCoord *co1, class CCoord *co2, class CCoord *co3 ) const;
};