//***************************************************************************//
//* libtrivia: various little routines for C++                              *//
//* string_convs: conversions to and from string                            *//
//* (C) Joachim Wuttke 2001, 2012-                                          *//
//* http://apps.jcns.fz-juelich.de                                          *//
//***************************************************************************//

#include <cmath>
#include <cstring>

#include <string>
#include <vector>
#include <iostream>

#include <boost/format.hpp>

#include "string_convs.hpp"

using std::string;
using std::cout;
using boost::format;

//***************************************************************************//
//*  Conversions to string                                                  *//
//***************************************************************************//

//! Returns input string.

string triv::strg( string val ) {
    return val;
}

//! Returns input C string as stdc++ string.

string triv::strg( const char* val ) {
    return val;
}

//! Convert bool to "y" or "n".

string triv::strg( bool val ) {
    return val ? "y" : "n";
}

//! Convert double to string.

string triv::strg( double val )
{
    string ret = str( format( "%g" ) % val );
    if( ret.find_first_of(".eE")==string::npos )
        ret += '.';
    return ret;
}

//! Convert int to string.

string triv::strg( int val )
{
    return str( format( "%i" ) % val );
}

//! Convert size_t to string.

string triv::strg( std::size_t val )
{
    return str( format( "%i" ) % ((int)val) );
}

//! Convert char to string.

string triv::strg( char val )
{
    return string(1, val);
}

//! Return time in SPHERES standard form.

string triv::time2strg( long seconds_since_the_epoch )
{
    char buf[40];
    struct tm tlocal;
    strftime( buf, 40, "%Y-%m-%i %a %H:%M:%S %Z",
              localtime_r(&seconds_since_the_epoch, &tlocal) );
    return buf;
}

//***************************************************************************//
//*  Convert from string                                                    *//
//***************************************************************************//

//! Try converting string to double.

bool triv::any2dbl( const string& inp, double *val )
{
    if( !inp.size() ) {
        cout << "expecting double, found empty input\n";
        return false;
    }
    if( inp=="-inf" || inp=="-oo" ) {
        *val = -INFINITY;
        return true;
    }
    if( inp=="inf" || inp=="oo" || inp=="+inf" || inp=="+oo" ) {
        *val = INFINITY;
        return true;
    }
    if(!strchr("+-0123456789.", inp[0])) {
        cout << "expecting double, found bad char[0] in [" << inp << "]\n";
        return false;
    }
    double mant; int expt;
    if (sscanf(inp.c_str(), "%lgm%i", &mant, &expt)==2) {
        *val = mant * pow(10.e0, (double)(-expt));
        return true;
    }
    if (sscanf(inp.c_str(), "%lg", val)==1)
        return true;
    cout << "expecting double, sscanf failed\n";
    return false;
}

//! Try converting string to integer.

bool triv::any2int( const string& inp, int *val )
{
    if(!inp.size()) {
        cout << "expecting integer, found empty input\n";
        return false;
    }
    if(!strchr("+-0123456789", inp[0])) {
        cout << "expecting integer, found bad char[0] in [" << inp << "]\n";
        return false;
    }
    if (sscanf(inp.c_str(), "%i", val)==1) return true;
    cout << "expecting integer, sscanf failed\n";
    return false;
}
